#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <unistd.h>
#include <sched.h>
#include <signal.h>

#include <sys/stat.h>
#include <sys/vfs.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>

typedef unsigned long long u64;
typedef signed long long   s64;
typedef unsigned int	   u32;
typedef signed int	   s32;
typedef unsigned short	   u16;
typedef signed short	   s16;
typedef unsigned char	   u8;
typedef signed char	   s8;

//#define DEBUG

#ifdef DEBUG
#define dprintf(x...)	printf(x)
#else
#define dprintf(x...)
#endif

#ifndef CGROUP_SUPER_MAGIC
#define CGROUP_SUPER_MAGIC 0x27e0eb
#endif

#define CHECK_TIME		3
#define GROUP_MAX_TASK_NUM	16

#define CFS_QUOTA	"/cpu.cfs_quota_us"
#define CFS_PERIOD	"/cpu.cfs_period_us"
#define CFS_TASKS	"/tasks"

static int check_cgroup_path(const char *group)
{
	struct stat st_group;
	struct statfs st_fs;
	
	if (stat(group, &st_group) < 0 || !S_ISDIR(st_group.st_mode))
		return -1;

	if (statfs(group, &st_fs) < 0 || st_fs.f_type != CGROUP_SUPER_MAGIC)
		return -1;

	return 0;
}

static int check_parameters(const char *group, int task_num, int nice,
			    const char *cpu_bind)
{
	if (check_cgroup_path(group) < 0)
		return -1;

	if (task_num < 0 || task_num > GROUP_MAX_TASK_NUM)
		return -1;

	if (nice > 19 || nice < -20)
		return -1;

	if (strcmp(cpu_bind, "no") && strcmp(cpu_bind, "same") &&
		strcmp(cpu_bind, "separate"))
		return -1;

	return 0;
}

static void write_value_to_file(const char *file, s64 value)
{
	FILE *fp;
	int ret;

	fp = fopen(file, "w+");
	if (!fp)
		printf("file: %s\n", file);
	assert(fp);
	ret = fprintf(fp, "%lld", value);
	assert(ret >= 0);
	fclose(fp);
}

static void group_add_task(const char *group, s64 pid)
{
	char path[256];

	strcpy(path, group);
	strcat(path, CFS_TASKS);
	write_value_to_file(path, pid);
}

static void do_hard_work(void)
{
	u64 s = 0;

	while (1) {
		s++;
		asm volatile("" ::: "memory");
		write_value_to_file("/dev/null", s);
	}
}

static void create_add_tasks(const char *group, int task_num, int nice,
			     const char *cpu_bind)
{
	int ret, cpu = 0, nr_cpu = sysconf(_SC_NPROCESSORS_CONF);
	
	while (task_num-- > 0) {
		s64 pid;
		cpu_set_t mask;

		pid = fork();
		assert(pid >= 0);
		if (!pid)
			do_hard_work();
		else {
			group_add_task(group, pid);
			setpriority(PRIO_PROCESS, pid, nice);
			if (!strcmp(cpu_bind, "no"))
				continue;
			if (!strcmp(cpu_bind, "separate")) {
				cpu++;
				cpu %= nr_cpu;
			}
			dprintf("bind task to cpu %d\n", cpu);
			CPU_ZERO(&mask);
			CPU_SET(cpu, &mask);
			ret = sched_setaffinity(pid, sizeof(mask), &mask);
			assert(ret==0);
		}
	}
}

static s64 read_value_from_file(const char *file)
{
	FILE *fp;
	s64 value;
	int ret;

	fp = fopen(file, "r");
	assert(fp);

	ret = fscanf(fp, "%lld", &value);
	assert(ret >= 0);
	fclose(fp);

	dprintf("read:%s value:%lld\n", file, value);
	return value;
}

static s64 read_cfs_quota(const char *group)
{
	char path[256];

	strcpy(path, group);
	strcat(path, CFS_QUOTA);
	return read_value_from_file(path);
}

static s64 read_cfs_period(const char *group)
{
	char path[256];

	strcpy(path, group);
	strcat(path, CFS_PERIOD);
	return read_value_from_file(path);
}

#define for_each_task(i, tasks)		\
		for (i = 0; i < GROUP_MAX_TASK_NUM && tasks[i]; i++)

static void read_group_tasks(const char *group, u64 tasks[GROUP_MAX_TASK_NUM])
{
	FILE *fp;
	char string[256];
	int ret, i = 0;

	strcpy(string, group);
	strcat(string, CFS_TASKS);

	fp = fopen(string, "r");
	assert(fp);

	while (fgets(string, 256, fp) != NULL) {
		ret = sscanf(string, "%lld", &tasks[i++]);
		assert(ret >= 0 && i < GROUP_MAX_TASK_NUM);
	}
	fclose(fp);

	dprintf("read tasks:\n");
	for_each_task(i, tasks)
		dprintf("---------task[%d]: %lld\n", i, tasks[i]);
	dprintf("end read tasks\n");
}

static u64 get_task_runtime(u64 pid)
{
	char filename [] = "/proc/1234567890/sched", string[256];
	u64 msecs, nsecs, total = 0;
	FILE *fp;
	int ret;

	sprintf(filename, "/proc/%lld/sched", pid);
	fp = fopen(filename, "r");
	assert(fp);
	 
	while (fgets(string, 256, fp) != NULL) {
		ret = sscanf(string, "se.sum_exec_runtime : %lld.%06lld\n",
			     &msecs, &nsecs);
		if (ret == 2) {
			total = msecs * 1000ULL + nsecs  / 1000ULL;
			dprintf("------msecs:%lld, nsecs:%lld total:%lld\n",
				msecs, nsecs, total);
			break;
		}
	}

	fclose(fp);
	return total;
}

static u64 get_tasks_runtime(u64 tasks[GROUP_MAX_TASK_NUM])
{
	u64 runtime = 0;
	int i;
	
	for_each_task(i, tasks)
		runtime += get_task_runtime(tasks[i]);
	return runtime;
}

static u64 get_usecs(void)
{
	struct timeval ts;

	gettimeofday(&ts, NULL);

	return ts.tv_sec * 1000000ULL + ts.tv_usec;
}

static int is_valid_rutime(u64 runtime, u64 expect_time)
{
	u64 diff;

	if (runtime < expect_time)
		return 1;
	diff = runtime - expect_time;
	if (diff < expect_time * 0.1)
		return 1;

	return 0;
}

static void usage(void)
{
	printf("\nUsage:  group-name task-number tasks-nice "
	       "cpu-bind(no,same,separate)\n");
}

int main(int argc, const char *argv[])
{
	const char *group, *cpu_bind;
	s64 quota, period;
	u64 tasks[GROUP_MAX_TASK_NUM], rstart, rend, cstart, cend, runtime,
		expect_runtime;
	int sleep_time, i, task_num, nice, ret = 0, checktime = CHECK_TIME;

	if (argc < 5) {
		usage();
		return -1;
	}

	group = argv[1];
	task_num = atoi(argv[2]);
	nice = atoi(argv[3]);
	cpu_bind = argv[4];

	if (check_parameters(group, task_num, nice, cpu_bind) < 0) {
		usage();
		return -1;
	}

	quota = read_cfs_quota(group);
	period = read_cfs_period(group);

	if (quota == -1)
		return 0;

	create_add_tasks(group, task_num, nice, cpu_bind);
	memset(tasks, 0, sizeof(tasks));
	read_group_tasks(group, tasks);

	sleep_time = (period * 3) /1000000;
	if (!sleep_time)
		sleep_time = 1;
	dprintf("sleep time: %d\n", sleep_time);

	while (checktime--) {
		cstart = get_usecs();
		rstart = get_tasks_runtime(tasks);
		sleep(sleep_time);
		rend = get_tasks_runtime(tasks);
		cend = get_usecs();

		runtime = rend - rstart;
		expect_runtime = ((float)(cend - cstart) / (float)period);
		expect_runtime *= quota;
		dprintf("begain-runtime:%lld end-runtime:%lld\n", rstart, rend);
		dprintf("expect:%lld, runtime:%lld\n", expect_runtime, runtime);

		if (!runtime)
			printf("Warning: the taks in the group seems not "
			       "be scheduled form %lld usecs\n",
				cend - cstart);

		if (!is_valid_rutime(runtime, expect_runtime)) {
			printf("runtime is longer than the limit: "
			       "runtime:%lld limit:%lld\n",
				runtime, expect_runtime);
			ret = -1;
			goto exit;
		}
	}

exit:
	for_each_task(i, tasks) {
		kill(tasks[i], SIGKILL);
		waitpid(tasks[i], NULL, 0);
	}

	return ret;
}
